#include "MSPd.h"

#define MAXSTORE (128)
#define OBJECT_NAME "octosplit~"

/* Pd version of octosplit~ */

static t_class *octosplit_class;

typedef struct _octosplit
{
    t_object x_obj;
    t_float x_f;
    int N;
    int N2;
    int lastN;
    int lastN2;
    void *list_outlet;
    void *phase_outlet;
    t_atom *list_data;
    int *bin_tmp;
    // Avoid buffer sharing overwrites with local copies
    t_float *inmag_loc;
    t_float *inphase_loc;
    t_float *t_offset_loc;
    t_float *b_offset_loc;
    t_float *manual_control_loc;
    int ramp_frames;
    int frames_left;
    float frame_duration;
    int vector_size;
    
    int table_offset;
    int bin_offset;
    float *last_mag;
    float *current_mag;
    int *last_binsplit;
    int *current_binsplit;
    int **stored_binsplits;
    short *stored_slots;
    short new_distribution;
    short interpolation_completed;
    short bypass;
    short initialize;
    short manual_override;
    long countdown_samps; // samps for a given fadetime
    long counter;
    int overlap_factor; // compensate for overlap in fade
    float sr;
    float fl_phase; // show phase as float
    int hopsamps; // number of samples to hop
    t_clock *phase_clock;
} t_octosplit;

void *octosplit_new(t_symbol *s, int argc, t_atom *argv);
t_int *offset_perform(t_int *w);
t_int *octosplit_perform(t_int *w);
void octosplit_dsp(t_octosplit *x, t_signal **sp);
void octosplit_showstate( t_octosplit *x );
void octosplit_bypass( t_octosplit *x, long toggle );
void octosplit_manual_override( t_octosplit *x, long toggle );
void octosplit_setstate (t_octosplit *x, t_symbol *msg, short argc, t_atom *argv);
void octosplit_ramptime (t_octosplit *x, t_symbol *msg, short argc, t_atom *argv);
int rand_index( int max);
void octosplit_scramble (t_octosplit *x);
void octosplit_spiral(t_octosplit *x);
void octosplit_squantize(t_octosplit *x, t_floatarg blockbins);
void octosplit_overlap( t_octosplit *x, long factor );

void octosplit_store( t_octosplit *x, int location );
void octosplit_recall( t_octosplit *x, int location );
void splitspeci( int *current_binsplit, int *last_binsplit, int bin_offset, int table_offset,
                float *current_mag, float *last_mag, float *inmag, float *dest_mag, int start, int end, int n,
                float oldfrac, float newfrac );
void splitspec( int *binsplit, int bin_offset, int table_offset,
               float *inmag, float *dest_mag, int start, int end, int n );

void octosplit_dsp_free( t_octosplit *x );

void octosplit_phaseout(t_octosplit *x);

void octosplit_tilde_setup(void){
    octosplit_class = class_new(gensym("octosplit~"), (t_newmethod)octosplit_new,
                                (t_method)octosplit_dsp_free, sizeof(t_octosplit),0,A_GIMME,0);
	CLASS_MAINSIGNALIN(octosplit_class, t_octosplit, x_f);
    class_addmethod(octosplit_class, (t_method)octosplit_dsp, gensym("dsp"),0);
    class_addmethod(octosplit_class, (t_method)octosplit_showstate, gensym("showstate"),0);
    class_addmethod(octosplit_class, (t_method)octosplit_scramble, gensym("scramble"),0);
    class_addmethod(octosplit_class, (t_method)octosplit_spiral, gensym("spiral"),0);
    class_addmethod(octosplit_class, (t_method)octosplit_squantize, gensym("squantize"),A_FLOAT,0);
    class_addmethod(octosplit_class, (t_method)octosplit_manual_override, gensym("manual_override"),A_FLOAT,0);
    class_addmethod(octosplit_class, (t_method)octosplit_store, gensym("store"),A_FLOAT,0);
    class_addmethod(octosplit_class, (t_method)octosplit_recall, gensym("recall"),A_FLOAT,0);
    class_addmethod(octosplit_class, (t_method)octosplit_overlap, gensym("overlap"),A_FLOAT,0);
    class_addmethod(octosplit_class, (t_method)octosplit_setstate, gensym("setstate"),A_GIMME,0);
    class_addmethod(octosplit_class, (t_method)octosplit_ramptime, gensym("ramptime"),A_GIMME,0);
    potpourri_announce(OBJECT_NAME);
}

void octosplit_phaseout(t_octosplit *x)
{
    outlet_float(x->phase_outlet, x->fl_phase);
    // post("%f", x->fl_phase);
}

void octosplit_overlap( t_octosplit *x, long overlap )
{
	if( overlap < 2 ){
		post("octosplit~: illegal overlap %d",overlap);
	}
	x->overlap_factor = overlap;
}

void octosplit_spiral(t_octosplit *x)
{
    int i;
    int offset1;
    int offset2;
    int offset3;
    int offset4;
    int offset5;
    int offset6;
    int offset7;
    
    int *current_binsplit = x->current_binsplit;
    int *last_binsplit = x->last_binsplit;
    
    int N2 = x->N2;
    
    x->new_distribution = 1;
    x->interpolation_completed = 0;
    
    offset1 = N2 / 8;
    offset2 = N2 / 4;
    offset3 = 3 * N2 / 8;
    offset4 = N2 / 4;
    offset5 = 5 * N2 / 8;
    offset6 = 3 * N2 / 4;
    offset7 = 7 * N2 / 8;
    
    for( i = 0; i < N2; i++ ){
        last_binsplit[i] = current_binsplit[i];
    }
    
    for( i = 0; i < N2 / 8; i += 8 ){
        current_binsplit[i] = i;
        current_binsplit[i + offset1] = i + 1;
        current_binsplit[i + offset2] = i + 2;
        current_binsplit[i + offset3] = i + 3;
        current_binsplit[i + offset4] = i + 4;
        current_binsplit[i + offset5] = i + 5;
        current_binsplit[i + offset6] = i + 6;
        current_binsplit[i + offset7] = i + 7;
    }
    x->frames_left = x->ramp_frames;
    if(! x->ramp_frames) { // Ramp Off - Immediately set last to current
        for( i = 0; i < N2; i++ ){
            last_binsplit[ i ] = current_binsplit[ i ];
        }
    }
}

void octosplit_squantize(t_octosplit *x, t_floatarg bb)
{
    int i, j, k;
    
    int maxblock;
    int iterations;
//    float fiterations;
    int bincount = 0;
//    int offset = 0;
    int subdiv = 8;
//    int binchunk;
    int *current_binsplit = x->current_binsplit;
    int *last_binsplit = x->last_binsplit;
//    int *bin_tmp = x->bin_tmp;
    int blockbins = (int) bb;
    int N2 = x->N2;
    
    maxblock = N2 / 8;
    if( blockbins < 1 || blockbins > maxblock ){
        error("%d is out of bounds - must be between 1 and %d", blockbins, maxblock);
        return;
    }
    
    iterations = N2 /  subdiv /  blockbins;
    // post("%d iterations", iterations);
    x->new_distribution = 1;
    x->interpolation_completed = 0;
    
    
    for( i = 0; i < N2; i++ ){
        last_binsplit[i] = current_binsplit[i];
    }
    
    if( iterations == 1 ){
        for( i = 0; i < N2 ; i++ ){
            current_binsplit[i] = i;
        }
    }
    else {
        for( k = 0; k < iterations; k++ ) {
            for( i = 0; i < N2; i += maxblock  ){
                for( j = 0; j < blockbins; j++ ){
                    if( i + j + k * blockbins < N2 ){
                        current_binsplit[i + j + k * blockbins] = bincount++;
                        // post("assigning %d to position %d", bincount-1, i+j+k*blockbins);
                    } else {
                        // error("%d out of range", i + j + k * blockbins);
                    }
                }
                
            }
        }
    }
    
    
    x->frames_left = x->ramp_frames;
    if(! x->ramp_frames) { // Ramp Off - Immediately set last to current
        for( i = 0; i < N2; i++ ){
            last_binsplit[ i ] = current_binsplit[ i ];
        }
    }
    
}

void octosplit_bypass( t_octosplit *x, long toggle )
{
    x->bypass = toggle;
    post("bypass %d", toggle);
}

void octosplit_manual_override( t_octosplit *x, long toggle )
{
    x->manual_override = toggle;
}

void octosplit_dsp_free( t_octosplit *x ){
    int i;
    
    if( ! x->initialize ){
        t_freebytes( x->list_data, (x->N + 2) * sizeof(t_atom) ) ;
        t_freebytes( x->last_binsplit, x->N2 * sizeof(int) );
        t_freebytes( x->current_binsplit, x->N2 * sizeof(int) );
        t_freebytes( x->bin_tmp , x->N2 * sizeof(int)) ;
        t_freebytes( x->last_mag , x->N2 * sizeof(float)) ;
        t_freebytes( x->current_mag, x->N2 * sizeof(float)) ;
        t_freebytes( x->stored_slots , x->N2 * sizeof(short));
        for( i = 0; i < MAXSTORE; i++ ){
            t_freebytes( x->stored_binsplits[i], x->N2 * sizeof(int));
        }
        t_freebytes( x->stored_binsplits , MAXSTORE * sizeof(int *));
        
        t_freebytes(x->inmag_loc, 8192 * sizeof(t_float));
        t_freebytes(x->inphase_loc, 8192 * sizeof(t_float));
        t_freebytes(x->t_offset_loc, 8192 * sizeof(t_float));
        t_freebytes(x->b_offset_loc, 8192 * sizeof(t_float));
        t_freebytes(x->manual_control_loc, 8192 * sizeof(t_float));
    }
}

void splitspeci( int *current_binsplit, int *last_binsplit, int bin_offset, int table_offset,
                float *current_mag, float *last_mag, float *inmag, float *dest_mag, int start, int end, int n,
                float oldfrac, float newfrac )
{
    int i;
    int bindex;
    
    for( i = 0; i < n; i++ ){
        last_mag[i] = current_mag[i] = 0.0;
    }
    
    for( i = start; i < end; i++ ){
        bindex = current_binsplit[ (i + table_offset) % n ];
        bindex = ( bindex + bin_offset ) % n;
        current_mag[ bindex ] = inmag[ bindex ];
        bindex = last_binsplit[ (i + table_offset) % n ];
        bindex = ( bindex + bin_offset ) % n;
        last_mag[ bindex ] = inmag[ bindex ];
    }
    for( i = 0; i < n; i++){
        if(! current_mag[i] && ! last_mag[i]){
            dest_mag[i] = 0.0;
        } else {
            dest_mag[i] = oldfrac * last_mag[i] + newfrac * current_mag[i];
        }
    }
    
}

void splitspec( int *binsplit, int bin_offset, int table_offset,
               float *inmag, float *dest_mag, int start, int end, int n )
{
    int i;
    int bindex;
    
    for( i = start; i < end; i++){
        bindex = binsplit[ (i + table_offset) % n ];
        bindex = ( bindex + bin_offset ) % n;
        dest_mag[ bindex ] = inmag[ bindex ];
    }
}


void octosplit_store( t_octosplit *x, int location )
{
    int **stored_binsplits = x->stored_binsplits;
    int *current_binsplit = x->current_binsplit;
    short *stored_slots = x->stored_slots;
    
    int i;
    
    if( location < 0 || location > MAXSTORE - 1 ){
        error("location must be between 0 and %d, but was %d", MAXSTORE, location);
        return;
    }
    for(i = 0; i < x->N2; i++ ){
        stored_binsplits[location][i] = current_binsplit[i];
    }
    stored_slots[location] = 1;
    
    post("stored bin split at location %d", location);
}

void octosplit_recall( t_octosplit *x, int location )
{
    int **stored_binsplits = x->stored_binsplits;
    int *current_binsplit = x->current_binsplit;
    int *last_binsplit = x->last_binsplit;
    short *stored_slots = x->stored_slots;
    int i;
    
    if( location < 0 || location > MAXSTORE - 1 ){
        error("location must be between 0 and %d, but was %d", MAXSTORE, location);
        return;
    }
    if( ! stored_slots[location] ){
        error("nothing stored at location %d", location);
        return;
    }
    
    for(i = 0; i < x->N2; i++ ){
        last_binsplit[i] = current_binsplit[i];
        current_binsplit[i] = stored_binsplits[location][i];
    }
    
    x->new_distribution = 1;
    x->interpolation_completed = 0;
    x->frames_left = x->ramp_frames;
    if(! x->ramp_frames) { // Ramp Off - Immediately set last to current
        for( i = 0; i < x->N2; i++ ){
            x->last_binsplit[ i ] = x->current_binsplit[ i ];
        }
    }
    
}

void *octosplit_new(t_symbol *s, int argc, t_atom *argv)
{
    int i;
    t_octosplit *x = (t_octosplit *)pd_new(octosplit_class);
    
    
    srand( time( 0 ) );
    
    for(i=0; i < 4; i++){
        inlet_new(&x->x_obj, &x->x_obj.ob_pd, gensym("signal"),gensym("signal"));
    }
    for(i=0; i < 16; i++){
        outlet_new(&x->x_obj, gensym("signal"));
    }
    x->list_outlet = (t_outlet *) outlet_new(&x->x_obj, gensym("list")); // to report random distribution
    x->phase_outlet = (t_outlet *) outlet_new((t_object *)x, &s_float);
    x->phase_clock = (t_clock *) clock_new((void *)x, (t_method)octosplit_phaseout);
    x->bypass = 0;
    x->table_offset = 0;
    x->bin_offset = 0;
    
    x->sr = sys_getsr();
 	x->counter = 0;
    x->overlap_factor = 8; // default
    
	x->countdown_samps = 1.0 * x->sr; // 1 second fade time by default
    
    x->initialize = 1;
    x->manual_override = 0;
    
    x->inmag_loc = (t_float *) t_getbytes(8192 * sizeof(t_float));
    x->inphase_loc = (t_float *) t_getbytes(8192 * sizeof(t_float));
    x->t_offset_loc = (t_float *) t_getbytes(8192 * sizeof(t_float));
    x->b_offset_loc = (t_float *) t_getbytes(8192 * sizeof(t_float));
    x->manual_control_loc = (t_float *) t_getbytes(8192 * sizeof(t_float));
    outlet_float(x->phase_outlet,0.666);
    return x;
}

t_int *octosplit_perform(t_int *w)
{
	
    int i;

    t_octosplit *x = (t_octosplit *) (w[1]);
    
    float *inmag = (t_float *)(w[2]);
    float *inphase = (t_float *)(w[3]);
    float *t_offset = (t_float *)(w[4]);
    float *b_offset = (t_float *)(w[5]);
    float *manual_control = (t_float *)(w[6]);
    
    float *outmag1 = (t_float *)(w[7]);
    float *outphase1 = (t_float *)(w[8]);
    float *outmag2 = (t_float *)(w[9]);
    float *outphase2 = (t_float *)(w[10]);
    float *outmag3 = (t_float *)(w[11]);
    float *outphase3 = (t_float *)(w[12]);
    float *outmag4 = (t_float *)(w[13]);
    float *outphase4 = (t_float *)(w[14]);
    
    float *outmag5 = (t_float *)(w[15]);
    float *outphase5 = (t_float *)(w[16]);
    float *outmag6 = (t_float *)(w[17]);
    float *outphase6 = (t_float *)(w[18]);
    float *outmag7 = (t_float *)(w[19]);
    float *outphase7 = (t_float *)(w[20]);
    float *outmag8 = (t_float *)(w[21]);
    float *outphase8 = (t_float *)(w[22]);
    
    //   float *sync = (t_float *)(w[23]);
    int n = (int)(w[23]);
    
    int table_offset = x->table_offset;
    int bin_offset = x->bin_offset;
    
    int *current_binsplit = x->current_binsplit;
    int *last_binsplit = x->last_binsplit;
    float *last_mag = x->last_mag;
    float *current_mag = x->current_mag;
    long counter = x->counter;
    long countdown_samps = x->countdown_samps;
    float frac, oldgain, newgain;
    t_float *inmag_loc = x->inmag_loc;
    t_float *inphase_loc = x->inphase_loc;
    t_float *t_offset_loc = x->t_offset_loc;
    t_float *b_offset_loc = x->b_offset_loc;
    t_float *manual_control_loc = x->manual_control_loc;
    int next_ptr = 24;
    int hopsamps = x->hopsamps;
    /****/
    
    // copy inputs to local buffers
    
    for(i = 0; i < n; i++){
        inmag_loc[i] = inmag[i];
        inphase_loc[i] = inphase[i];
        t_offset_loc[i] = t_offset[i];
        b_offset_loc[i] = b_offset[i];
        manual_control_loc[i] = manual_control[i];
    }
    
    table_offset = *t_offset_loc * n;
    bin_offset = *b_offset_loc * n;
    
    if( table_offset  < 0 )
        table_offset *= -1;
    if( bin_offset  < 0 )
        bin_offset *= -1;
  	
    
    // n == fftsize / 2 (N2)
    // n is the number of "bins", and is also the number of values in each signal vector
    
    if( x->bypass ){
        for( i = 0; i < n; i++){
            outmag1[i] = outmag2[i] = outmag3[i] = outmag4[i] =
            outmag5[i] = outmag6[i] = outmag7[i] = outmag8[i] = inmag_loc[i] * 0.5; // scale down for bypass
            outphase1[i] = outphase2[i] = outphase3[i] = outphase4[i] =
            outphase5[i] = outphase6[i] = outphase7[i] = outphase8[i] = inphase_loc[i];
        }
        return (w + next_ptr);
    }
    
    // ZERO OUT MAGNITUDES AND COPY PHASE TO OUTPUT
    for( i = 0; i < n; i++ ){
        outmag1[i] = outmag2[i] = outmag3[i] = outmag4[i] =
        outmag5[i] = outmag6[i] = outmag7[i] = outmag8[i] = 0.0;
        outphase1[i] = outphase2[i] = outphase3[i] = outphase4[i] =
        outphase5[i] = outphase6[i] = outphase7[i] = outphase8[i] = inphase_loc[i];
    }
    
    // Special case of live control over interpolation
    if( x->manual_override ){
        
        // do interpolation
        frac = *manual_control_loc ;
        // sanity check here
        if( frac < 0 )
            frac = 0;
        if( frac >1.0 )
            frac = 1.0;
        oldgain = cos( frac * PIOVERTWO );
        newgain = sin( frac * PIOVERTWO );
        
        
        splitspeci( current_binsplit, last_binsplit, bin_offset, table_offset,
                   current_mag, last_mag, inmag_loc, outmag1, 0, n/8, n, oldgain, newgain );
        
        splitspeci( current_binsplit, last_binsplit, bin_offset, table_offset,
                   current_mag, last_mag, inmag_loc, outmag2, n/8, n/4, n, oldgain, newgain );
        
        splitspeci( current_binsplit, last_binsplit, bin_offset, table_offset,
                   current_mag, last_mag, inmag_loc, outmag3, n/4, 3*n/8, n, oldgain, newgain );
        
        splitspeci( current_binsplit, last_binsplit, bin_offset, table_offset,
                   current_mag, last_mag, inmag_loc, outmag4, 3*n/8, n/2, n, oldgain, newgain );
        
        splitspeci( current_binsplit, last_binsplit, bin_offset, table_offset,
                   current_mag, last_mag, inmag_loc, outmag5, n/2, 5*n/8, n, oldgain, newgain );
        
        splitspeci( current_binsplit, last_binsplit, bin_offset, table_offset,
                   current_mag, last_mag, inmag_loc, outmag6, 5*n/8, 3*n/4, n, oldgain, newgain );
        
        splitspeci( current_binsplit, last_binsplit, bin_offset, table_offset,
                   current_mag, last_mag, inmag_loc, outmag7, 3*n/4, 7*n/8, n, oldgain, newgain );
        
        splitspeci( current_binsplit, last_binsplit, bin_offset, table_offset,
                   current_mag, last_mag, inmag_loc, outmag8, 7*n/8, n, n, oldgain, newgain );
        
        
        return (w + next_ptr);
    }
    
    // Normal operation
    if( x->new_distribution ){
        x->new_distribution = 0;
        // put out contents of last distribution
        splitspec(last_binsplit, bin_offset, table_offset, inmag_loc, outmag1, 0, n/8, n);
        splitspec(last_binsplit, bin_offset, table_offset, inmag_loc, outmag2, n/8, n/4, n);
        splitspec(last_binsplit, bin_offset, table_offset, inmag_loc, outmag3, n/4, 3*n/8, n);
        splitspec(last_binsplit, bin_offset, table_offset, inmag_loc, outmag4, 3*n/8, n/2, n);
        splitspec(last_binsplit, bin_offset, table_offset, inmag_loc, outmag5, n/2, 5*n/8, n);
        splitspec(last_binsplit, bin_offset, table_offset, inmag_loc, outmag6, 5*n/8, 3*n/4, n);
        splitspec(last_binsplit, bin_offset, table_offset, inmag_loc, outmag7, 3*n/4, 7*n/8, n);
        splitspec(last_binsplit, bin_offset, table_offset, inmag_loc, outmag8, 7*n/8, n, n);
        frac = 0.0;
        
    } else if ( x->interpolation_completed ) {
        // put out contents of current distribution
        splitspec(current_binsplit, bin_offset, table_offset, inmag_loc, outmag1, 0, n/8, n);
        splitspec(current_binsplit, bin_offset, table_offset, inmag_loc, outmag2, n/8, n/4, n);
        splitspec(current_binsplit, bin_offset, table_offset, inmag_loc, outmag3, n/4, 3*n/8, n);
        splitspec(current_binsplit, bin_offset, table_offset, inmag_loc, outmag4, 3*n/8, n/2, n);
        splitspec(current_binsplit, bin_offset, table_offset, inmag_loc, outmag5, n/2, 5*n/8, n);
        splitspec(current_binsplit, bin_offset, table_offset, inmag_loc, outmag6, 5*n/8, 3*n/4, n);
        splitspec(current_binsplit, bin_offset, table_offset, inmag_loc, outmag7, 3*n/4, 7*n/8, n);
        splitspec(current_binsplit, bin_offset, table_offset, inmag_loc, outmag8, 7*n/8, n, n);
        frac = 1.0;
    } else {
        // do interpolation
        frac = (float) counter / (float) countdown_samps;
        
        oldgain = cos( frac * PIOVERTWO );
        newgain = sin( frac * PIOVERTWO );
        
        splitspeci( current_binsplit, last_binsplit, bin_offset, table_offset,
                   current_mag, last_mag, inmag_loc, outmag1, 0, n/8, n, oldgain, newgain );
        splitspeci( current_binsplit, last_binsplit, bin_offset, table_offset,
                   current_mag, last_mag, inmag_loc, outmag2, n/8, n/4, n, oldgain, newgain );
        splitspeci( current_binsplit, last_binsplit, bin_offset, table_offset,
                   current_mag, last_mag, inmag_loc, outmag3, n/4, 3*n/8, n, oldgain, newgain );
        splitspeci( current_binsplit, last_binsplit, bin_offset, table_offset,
                   current_mag, last_mag, inmag_loc, outmag4, 3*n/8, n/2, n, oldgain, newgain );
        splitspeci( current_binsplit, last_binsplit, bin_offset, table_offset,
                   current_mag, last_mag, inmag_loc, outmag5, n/2, 5*n/8, n, oldgain, newgain );
        splitspeci( current_binsplit, last_binsplit, bin_offset, table_offset,
                   current_mag, last_mag, inmag_loc, outmag6, 5*n/8, 3*n/4, n, oldgain, newgain );
        splitspeci( current_binsplit, last_binsplit, bin_offset, table_offset,
                   current_mag, last_mag, inmag_loc, outmag7, 3*n/4, 7*n/8, n, oldgain, newgain );
        splitspeci( current_binsplit, last_binsplit, bin_offset, table_offset,
                   current_mag, last_mag, inmag_loc, outmag8, 7*n/8, n, n, oldgain, newgain );
        
        /* end of interpolation */
        
        
        counter += hopsamps;
        // counter += 128;
        if( counter >= countdown_samps )
        {
            counter = countdown_samps;
            x->interpolation_completed = 1;
        }
    }
    
    x->fl_phase = frac;
    clock_delay(x->phase_clock,0.0);
    x->counter = counter;
    return (w + next_ptr);
}


void octosplit_scramble (t_octosplit *x)
{
    int i, j;
    int used;
    
    int max = x->N2;
    int bindex;
    
    int *current_binsplit = x->current_binsplit;
    int *last_binsplit = x->last_binsplit;
    int *bin_tmp = x->bin_tmp;
    
    x->new_distribution = 1;
    x->interpolation_completed = 0;
    
    // post("scrambling");
    
    // Copy current mapping to last mapping (first time this will be all zeros)
    
    for( i = 0; i < x->N2; i++ ){
        
        last_binsplit[i] = current_binsplit[i];
    }
    
    
    for( i = 0; i < max; i++ ){
        bin_tmp[i] = i;
        
    }
    
    used = max;
    
    // This randomly distributes each bin number (to occur once each in a random location)
    
    for( i = 0; i < max; i++ ){
        bindex = rand_index( used );
        current_binsplit[i] = bin_tmp[bindex];
        for(j = bindex; j < used - 1; j++) {
            bin_tmp[j] = bin_tmp[j+1];
        }
        --used;
    }
    x->counter = 0;
    if(! x->countdown_samps ) { // Ramp Off - Immediately set last to current
        for( i = 0; i < x->N2; i++ ){
            x->last_binsplit[ i ] = x->current_binsplit[ i ];
        }
    }
}

int rand_index( int max) {
    int rand();
    return ( rand() % max );
}

void octosplit_setstate (t_octosplit *x, t_symbol *msg, short argc, t_atom *argv) {
    short i;
    
    if( argc != x->N2 ){
        error("list must be of length %d, but actually was %d", x->N2, argc);
        return;
    }
    for( i = 0; i < x->N2; i++ ){
        x->last_binsplit[ i ] = x->current_binsplit[ i ];
        x->current_binsplit[ i ] = 0;
    }
    for (i=0; i < argc; i++) {
        x->current_binsplit[i] = atom_getintarg(i, argc, argv );
		
    }
    x->frames_left = x->ramp_frames;
    if(! x->ramp_frames) { // Ramp Off - Immediately set last to current
        for( i = 0; i < x->N2; i++ ){
            x->last_binsplit[ i ] = x->current_binsplit[ i ];
        }
    }
    
    return;
}

void octosplit_ramptime (t_octosplit *x, t_symbol *msg, short argc, t_atom *argv) {
    float rampdur;
    
 	rampdur = atom_getfloatarg(0,argc,argv) * 0.001; // convert from milliseconds
 	x->countdown_samps = rampdur * x->sr;
 	x->counter = 0;
    //  post("countdown samps :%d", x->countdown_samps  );
}

// REPORT CURRENT SHUFFLE STATUS
void octosplit_showstate (t_octosplit *x ) {
    
    t_atom *list_data = x->list_data;
    
    short i, count;
    
    count = 0;
    // post("showing %d data points", x->N2);
    
    if(! x->initialize){
        for( i = 0; i < x->N2; i++ ) {
            SETFLOAT(list_data+count,(float)x->current_binsplit[i]);
            ++count;
        }
        outlet_list(x->list_outlet,0,x->N2,list_data);
    }
    return;
}

void octosplit_dsp(t_octosplit *x, t_signal **sp)
{
    int i;
    float R, funda;
    
    //  x->vector_size = sys_getblksize();
    x->vector_size = sys_getblksize();
    
    /*
     
     FFT size (N) is the twice the vector size.
     There is one vector with the magnitudes, and another with the phases, each of size N/2 (N2).
     Therefore N2 (or the vector size) is the actual number of bins.
     Therefore N2/split is the number of bins assigned to each splitted channel.
     
     */
    post("vector size %d, sys vector size: %d",x->vector_size, sys_getblksize() );
    post("octosplit: samples per vector: %d, sys blocksize %d", sp[0]->s_n, sys_getblksize());
    if( ! sp[0]->s_sr ){
        error("octosplit~: zero sample rate!");
        return;
    }
	
    
    if( x->initialize || x->sr != sys_getsr() || x->N != sp[0]->s_n){
        
        x->sr = sys_getsr();
        x->N = sp[0]->s_n;
        x->N2 = sp[0]->s_n / 2;
        post("FFT size is %d, N2 is %d",x->N, x->N2);
        R = sys_getsr();
        post("sampling rate: %f, vector thinks it is: %f", sys_getsr(), sp[0]->s_sr);
        funda = R / (2. * (float) x->N) ;
   /*
        x->list_data = t_getbytes( (x->N + 2) * sizeof(t_atom));
        x->last_binsplit = t_getbytes( x->N2 * sizeof(int));
        x->current_binsplit = t_getbytes( x->N2 * sizeof(int));
        x->bin_tmp = t_getbytes( x->N2 * sizeof(int)) ;
        x->last_mag = t_getbytes( x->N2 * sizeof(float)) ;
        x->current_mag = t_getbytes( x->N2 * sizeof(float)) ;
        x->stored_slots = t_getbytes( x->N2 * sizeof(short));
        x->stored_binsplits = t_getbytes( MAXSTORE * sizeof(int *));
        for( i = 0; i < MAXSTORE; i++ ){
            x->stored_binsplits[i] = t_getbytes( x->N2 * sizeof(int));
        }
        */
        
        if(x->initialize){
            x->list_data = (t_atom *) getbytes((x->N + 2) * sizeof(t_atom));
            x->last_binsplit = (int *) getbytes( x->N2 * sizeof(int));
            x->current_binsplit = (int *) getbytes( x->N2 * sizeof(int));
            x->bin_tmp = (int *) getbytes( x->N2 * sizeof(int)) ;
            x->last_mag = (float *) getbytes(x->N2 * sizeof(float)) ;
            x->current_mag = (float *) getbytes( x->N2 * sizeof(float)) ;
            x->stored_slots = (short *) getbytes( x->N2 * sizeof(short));
            x->stored_binsplits = (int **)  getbytes( MAXSTORE * sizeof(int *));
            for( i = 0; i < MAXSTORE; i++ ){
                x->stored_binsplits[i] = (int *)getbytes( x->N2 * sizeof(int));
            }
            x->lastN = x->N;
            x->lastN2 = x->N2;
        } else {
            x->list_data = (t_atom *) resizebytes((void *)x->list_data,(x->lastN + 2) * sizeof(t_atom),(x->N + 2) * sizeof(t_atom));
            x->last_binsplit = (int *) resizebytes((void *)x->last_binsplit,x->lastN2 * sizeof(int),x->N2 * sizeof(int));
            x->current_binsplit = (int *) resizebytes((void *)x->current_binsplit,x->lastN2 * sizeof(int),x->N2 * sizeof(int));
            x->bin_tmp = (int *) resizebytes((void *)x->current_binsplit,x->lastN2 * sizeof(int),x->N2 * sizeof(int));
            x->last_mag = (float *) resizebytes((void *)x->last_mag,x->lastN2 * sizeof(float),x->N2 * sizeof(float));
            x->current_mag = (float *) resizebytes((void *)x->current_mag,x->lastN2 * sizeof(float),x->N2 * sizeof(float));
            x->stored_slots = (short *) resizebytes((void *)x->stored_slots,x->lastN2 * sizeof(short),x->N2 * sizeof(short));
            for( i = 0; i < MAXSTORE; i++ ){
                x->stored_binsplits[i] = (int *) resizebytes((void *)x->stored_binsplits[i],x->lastN2 * sizeof(int),x->N2 * sizeof(int));
            }
        }
        
        x->frame_duration = (float) sp[0]->s_n / sp[0]->s_sr;
        octosplit_scramble( x );
        
        for( i = 0; i < x->N2; i++ ){
            x->last_binsplit[i] = x->current_binsplit[i];
        }
        x->initialize = 0;
        x->counter = 0;
        // post("ramp frames: %d", x->ramp_frames);
        
    }
    
    if(x->vector_size == 0) {
        post("zero vector size!");
        return;
    } else {
        x->hopsamps = x->N / x->overlap_factor;
        post("hop samps: %d, overlap: %d", x->hopsamps, x->overlap_factor);
    }
    dsp_add(octosplit_perform, 23, x, sp[0]->s_vec, sp[1]->s_vec, sp[2]->s_vec, sp[3]->s_vec,
            sp[4]->s_vec, sp[5]->s_vec,sp[6]->s_vec, sp[7]->s_vec, sp[8]->s_vec,sp[9]->s_vec,
            sp[10]->s_vec,sp[11]->s_vec, sp[12]->s_vec, sp[13]->s_vec,
            sp[14]->s_vec,sp[15]->s_vec, sp[16]->s_vec, sp[17]->s_vec,
            sp[18]->s_vec,sp[19]->s_vec, sp[20]->s_vec,
            sp[0]->s_n);
}


